from typing import Any, Optional, List, Union, Callable
from typing_extensions import Literal
import skema
import fastjsonschema
from jsonschema import validate
from prtty import prettify


class dictlike(dict):
    def __getattr__(self, name):
        try:
            return dict.__getitem__(self, name)
        except KeyError:
            return object.__getattribute__(self, name)

    @classmethod
    def from_dict(cls, obj: dict):
        assert isinstance(obj, dict)
        return cls(**obj)

    @classmethod
    def from_(cls, obj: dict):
        assert isinstance(obj, dict)
        return cls(**obj)
    
    from_dict = from_

    def validate(self):
        # return self.validate_(self)
        return validate(instance=self, schema=self._schema)

    @classmethod
    def fake(cls, resolvers={}):
        return cls(
            **(
                skema.fake_data(
                    cls._schema, amount=1, from_json=True, resolvers=resolvers
                )[0]
                or {}
            )
        )

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):
        return f"{self.__class__.__name__}({prettify(self)})"



from skema.parser import parse
from lark import Tree

def reconstruct(string) -> Tree:
    try:
        from graphql import build_schema, DocumentNode
        from .print_graphql_schema import print_schema
    except ImportError:
        raise Exception('need graphql-core package to reconstruct graphql')
    schema = build_schema(string)
    skema = print_schema(schema)
    # print(skema.replace(' ', '.'))
    return skema
    # return parse(skema)



from skema.reconstruct import from_graphql
from graphql import build_schema

def test_1():
    print()
    s = (
        """
    type X {
        a: String
        b: Int
    }
    interface Interface {
        a: String
    }
    interface Node {
        b: String
    }
    type Y implements Interface {
        a: String @direct
        b: X
    }
    union A = X | Y
    scalar Scalar

    directive @direct(
    reason: String = "No longer supported"
    ) on FIELD_DEFINITION | ENUM_VALUE

    type Listed {
        f: [X]
        j: String
    }

    input Input {
        a: Int
    }

    """
    )
    skema = from_graphql(s)
    print(skema)



import os
import coloredlogs
import logging

logger = logging.getLogger(__name__)

LEVEL = 'DEBUG' if os.getenv('DEBUG') else 'ERROR'

coloredlogs.DEFAULT_FIELD_STYLES = {'asctime': {'color': 'white'}, 'hostname': {'color': 'white'}, 'levelname': {
    'color': 'white', 'bold': True}, 'name': {'color': 'white'}, 'programname': {'color': 'white'}}
coloredlogs.install(
    fmt='%(asctime)s %(module)s %(levelname)s %(message)s', level=LEVEL, logger=logger)



from lark import Lark
from lark.indenter import Indenter
from lark.reconstruct import Reconstructor
from .logger import logger
from populate import indent_to

tree_grammar = r"""
    start: (_NL* root_pair)+ _NL*

    scalar: "Str" -> type_str
        | "Int" -> type_int
        | "Float" -> type_float
        | "Bool" -> type_bool
        | "Any" -> type_any
        | "null" -> literal_null
        | "true" -> literal_true
        | "false" -> literal_false
        | ESCAPED_STRING -> literal_string
        | SIGNED_INT -> literal_integer
        | NAME -> reference
        | /\/.*\// -> regex


    literal_ellipsis: "..."
    union: (value ("|" (scalar | list))+)
    intersection: value (("&" scalar)+ _NL | "&" _NL object)

    ?value: scalar
        | union

    _TRIPLE_QUOTE: "\"\"\""

    annotation: _TRIPLE_QUOTE _NL (/.+/ _NL)* _TRIPLE_QUOTE _NL

    root_pair: [annotation] NAME ":" (_NL object | value _NL | intersection | list _NL)
    required_pair: [annotation] NAME ":" (_NL object | intersection | value _NL | list _NL)
    optional_pair: [annotation] NAME "?:" (_NL object | intersection | value _NL | list _NL)

    list: "[" (_NL object | value) "]"

    object: _INDENT (required_pair | optional_pair)+ [literal_ellipsis _NL] _DEDENT

    COMMENT: /#.*/

    %import common.CNAME -> NAME
    %import common.WS_INLINE
    %import common.ESCAPED_STRING
    %import common.SIGNED_NUMBER
    %import common.INT
    %import common.FLOAT
    %import common.SIGNED_INT
    %import common.SIGNED_FLOAT
    %declare _INDENT _DEDENT
    %ignore WS_INLINE
    %ignore COMMENT

    
    _NL: ["\r"] "\n" " "*
"""


class TreeIndenter(Indenter):
    NL_type = "_NL"
    OPEN_PAREN_types = []
    CLOSE_PAREN_types = []
    INDENT_type = "_INDENT"
    DEDENT_type = "_DEDENT"
    tab_len = 4


parser = Lark(
    tree_grammar,
    parser="lalr",
    postlex=TreeIndenter(),
    lexer_callbacks={"COMMENT": lambda c: None},
)


def parse(string):
    string = "\n".join([l for l in string.split("\n") if l.strip()])
    string = indent_to('', string)
    logger.debug(string)
    t = parser.parse(string + "\n")
    logger.debug(t.pretty())
    return t



from prtty import pretty
from lark import Token
from ..transformers import Transformer
from funcy import merge, lmap, omit
from ..parser import parser

ELLIPSIS = "..."


class JsonSchema(Transformer):
    def __init__(self, ref=None):
        self.ref = ref

    def start(self, children):
        ref = self.ref or get_first_key(children[0])
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "$ref": "#/definitions/" + ref,
            "definitions": merge(*children),
        }

    def type_str(self, _):
        return {"type": "string"}

    def type_bool(self, _):
        return {"type": "boolean"}

    def type_int(self, _):
        return {"type": "number", "multipleOf": 1.0}

    def type_float(self, _):
        return {"type": "number"}

    def type_any(self, _):
        return {}

    def literal_null(self, _):
        return {"const": None}

    def literal_true(self, _):
        return {"const": True}

    def literal_false(self, _):
        return {"const": False}

    def literal_string(self, children):
        value, = children
        return {"const": value[1:-1]}

    def literal_integer(self, children):
        value, = children
        return {"const": int(value)}

    def literal_ellipsis(self, _):
        return ELLIPSIS

    def regex(self, children):
        value, = children
        return {"type": "string", "pattern": value[1:-1]}

    def reference(self, children):
        value, = children
        return {"$ref": f"#/definitions/{value}"}

    def annotation(self, children):
        value, = children
        return value

    # TODO remove and make range better
    def scalar(self, children):
        value, = children
        return value

    def bounded_range(self, children):
        are_floats = any(["." in x for x in children])
        if are_floats:
            children = lmap(float, children)
        else:
            children = lmap(int, children)
        low, high = children
        res = {"type": "number", "minimum": low, "maximum": high}
        if not are_floats:
            res.update({"multipleOf": 1.0})
        return res

    def low_bounded_range(self, children):
        are_floats = any(["." in x for x in children])
        if are_floats:
            children = lmap(float, children)
        else:
            children = lmap(int, children)
        low, = children
        res = {"type": "number", "minimum": low}
        if not are_floats:
            res.update({"multipleOf": 1.0})
        return res

    def high_bounded_range(self, children):
        are_floats = any(["." in x for x in children])
        if are_floats:
            children = lmap(float, children)
        else:
            children = lmap(int, children)
        high, = children
        res = {"type": "number", "maximum": high}
        if not are_floats:
            res.update({"multipleOf": 1.0})
        return res

    def object(self, children):
        required = [
            get_first_key(c) for c in children if c != ELLIPSIS and c["required"]
        ]
        children = lmap(lambda o: omit(o, ["required"]), children)
        properties = merge(*children)
        # if ELLIPSIS in children:
        #     children.remove(ELLIPSIS)
        #     if len(children) == 1:
        #         res = {"type": "object"}
        #     else:
        #         res = {"type": "object", "properties": properties}

        res = {
            "type": "object",
            "properties": properties,
            "additionalProperties": False,
            "required": required,
        }
        if required:
            res.update({"required": required})
        return res

    def list(self, children):
        value, = children
        return {"type": "array", "items": value}

    def union(self, children):
        if all(["const" in x for x in children]):
            return {"enum": children}
        return {"anyOf": children}

    def intersection(self, children):
        return {"allOf": children}

    def required_pair(self, children):
        key, value = children
        annotation = ""
        res = {str(key): {**value, "description": str(annotation)}, "required": True,}
        return res

    def root_pair(self, children):
        key, value = children
        annotation = ""
        res = {str(key): {"title": str(key), "description": str(annotation), **value}}
        return res

    def optional_pair(self, children):
        key, value = children
        res = {str(key): value, "required": False}
        return res

    pass


def get_first_key(obj):
    keys = list(obj.keys())
    return keys[0]



from prtty import pretty
from populate import indent_to, populate_string
from lark import Token, Tree, v_args
from funcy import merge, lmap, omit, concat
from ..parser import parser
from ..transformers import Transformer
from ..support import structure, types, composed_types, literals

ELLIPSIS = "..."


@v_args(tree=True)
class AddInitializersMetas(Transformer):
    def get_initializer(self, node: Tree):
        initializer = {
            types.STR: lambda: "str($value)",
            types.INT: lambda: "int($value)",
            types.FLOAT: lambda: "float($value)",
            types.BOOL: lambda: "bool($value)",
            types.ANY: lambda: "$value",
            literals.STRING: lambda: "$value",
            literals.TRUE: lambda: "True",
            literals.FALSE: lambda: "False",
            literals.INTEGER: lambda: "$value",
            literals.NULL: lambda: "None",
            structure.REFERENCE: lambda: f"{node.children[0]}.from_($value)",
            composed_types.OBJECT: lambda: f"unexpected object",
            composed_types.UNION: lambda: "$value",
            composed_types.LIST: lambda: f'[{self.get_initializer(node.children[0]).replace("$value", "x")} for x in $value]',
        }[str(node.data)]()
        return initializer

    def required_pair(self, t):
        k, v = t.children
        initializer = self.get_initializer(v)
        if not isinstance(t._meta, dict):
            t._meta = {}
        t._meta = {**t._meta, "initializer": initializer}
        return t

    optional_pair = required_pair


imports = """
from skema import dictlike
from typing import Optional, List, Any, Union
from typing_extensions import Literal

"""


class Python(Transformer):
    def __init__(self, ref=None):
        self.ref = ref

    def __default__(self, data, children, meta):
        raise NotImplementedError(f"{data} is not implemented in grahql") from None

    def start(self, children):
        return imports + "\n".join(children)

    def type_str(self, _):
        return "str"

    def type_bool(self, _):
        return "bool"

    def type_int(self, _):
        return "int"

    def type_float(self, _):
        return "float"

    def type_any(self, _):
        return "Any"

    def literal_null(self, _):
        return "None"

    def literal_true(self, _):
        return f"Literal[True]"

    def literal_false(self, _):
        return f"Literal[False]"

    def literal_string(self, children):
        value, = children
        return f"Literal[{value}]"

    def literal_number(self, children):
        value, = children
        return f"Literal[{value}]"

    def literal_ellipsis(self, _):
        return ELLIPSIS

    def regex(self, children):
        # value, = children
        return "str"

    def reference(self, children):
        value, = children
        return f'"{value}"'

    def annotation(self, children):
        value, = children
        return value

    # TODO remove and make range better
    def scalar(self, children):
        value, = children
        return value

    def object(self, children):
        types = "\n".join([x for x, _, _ in children]) + "\n"
        arguments = ",\n".join([x for _, x, _ in children])
        initializers = ",\n".join([x for _, _, x in children])
        template = indent_to(
            "",
            """
            class $key(dictlike):
                $annotation
                ${{indent_to('    ', types)}}
                
                def __init__(
                    self,
                    *, 
                    ${{indent_to('        ', arguments)}}, 
                    **kwargs
                ):
                    super().__init__(
                        ${{indent_to('            ', initializers)}},
                        **kwargs
                    )
            """,
        )
        return populate_string(
            template,
            dict(
                types=types,
                arguments=arguments,
                initializers=initializers,
                indent_to=indent_to,
            ),
        )

    def list(self, children):
        value, = children
        return f"List[{value}]"

    def union(self, children):
        return f'Union[{", ".join(children)}]'

    @v_args(meta=True)
    def required_pair(self, children, meta):
        k, v = children
        if "$key" in v:
            return v.replace("$key", k)
        return k + ": " + v, k, f'{k}={meta["initializer"].replace("$value", k)} if {k} != None else None'

    @v_args(meta=True)
    def optional_pair(self, children, meta):
        k, v = children
        if "$key" in v:
            return v.replace("$key", k)
        else:
            return (
                f"{k}: Optional[{v}] = None\n",
                f"{k}=None",
                f'{k}={meta["initializer"].replace("$value", k)} if {k} != None else None',
            )

    @v_args(meta=True)
    def root_pair(self, children, meta):
        k, v = children
        annotation = meta['annotation'] if 'annotation' in meta else ''
        annotation = '\n    '.join(annotation.split('\n'))
        annotation = annotation and f'"""\n    {annotation}\n    """\n'
        if "$key" in v:
            return v.replace("$key", k).replace("$annotation", annotation)
        else:
            return f"{k} = {v}\n{k}.from_ = lambda x: x\n"

    pass


def get_first_key(obj):
    keys = list(obj.keys())
    return keys[0]



from lark import Visitor, Tree, Token, v_args
from .support import Transformer
from functools import partial
from funcy import cat, flip, collecting
from prtty import pretty
from collections import defaultdict
from toposort import toposort, toposort_flatten
from ordered_set import OrderedSet
from ..types import UniqueKey
from ..support import structure, literals, composed_types
import uuid
from copy import copy



@v_args(tree=True)
class RemoveAnnotations(Transformer):
    def root_pair(self, t: Tree):
        first, *_ = t.children
        if isinstance(first, Tree) and first.data == structure.ANNOTATION:
            annotation = t.children.pop(0).children[0]
        else:
            annotation = ''
        meta = t.meta if isinstance(t.meta, dict) else {}
        t._meta = {**meta, 'annotation': str(annotation)}
        return t
    required_pair = root_pair
    optional_pair = root_pair



@v_args(tree=True)
class RemoveEllipses(Transformer):
    def object(self, t: Tree):
        last = t.children[-1]
        meta = t.meta if isinstance(t.meta, dict) else {}
        if isinstance(last, Tree) and last.data == literals.ELLIPSIS:
            t._meta = {**meta, 'ellipsis': True}
            t.children.pop(-1)
            return t
        t._meta = {**meta, 'ellipsis': False}
        return t



@v_args(tree=True)
class GetDependencies(Transformer):
    dependencies: defaultdict
    
    def __init__(self,):
        self.dependencies = defaultdict(OrderedSet)
        
    def start(self, tree):
        ordered_deps = list(toposort_flatten(self.dependencies_as_sets))
        for k in ordered_deps:
            v = self.dependencies[k]
            to_process = [x for x in v if x in self.dependencies]
            for x in to_process:
                self.dependencies[(k)].update(self.dependencies[x])
        return Tree("start", tree.children, meta={"dependencies": self.dependencies})

    @property
    def dependencies_as_sets(self,):
        return {k: set(v) for k, v in self.dependencies.items()}

    def required_pair(self, t):
        this, value = t.children
        if value.data == "object":
            for child in value.children:
                k, _ = child.children
                self.dependencies[UniqueKey(k)].add(UniqueKey(this))
        return t

    optional_pair = required_pair
    root_pair = required_pair


@v_args(tree=True)
class AddListMetas(Transformer):
    def required_pair(self, tree: Tree):
        name, list_node = tree.children
        if not list_node.data == "list":
            return tree
        list_node._meta = {**list_node._meta, "parent_key": name}
        return tree

    optional_pair = required_pair
    root_pair = required_pair


@v_args(tree=True)
class AddUnionMetas(Transformer):
    def required_pair(self, tree: Tree):
        name, list_node = tree.children
        if not list_node.data == "union":
            return tree
        list_node._meta = {**list_node._meta, "parent_key": name}
        return tree

    optional_pair = required_pair
    root_pair = required_pair



from lark import Visitor, Tree, Token, v_args
from lark.tree import Meta
from .support import Transformer
from ..support import capitalize, structure, composed_types, literals, types
from ..logger import logger
from functools import partial
from funcy import cat, flip, collecting
from prtty import pretty
from collections import defaultdict
from toposort import toposort, toposort_flatten
from ordered_set import OrderedSet
from ..types import UniqueKey
import uuid
from copy import copy


class TransformerWithMeta(Transformer):
    meta: dict

    def __init__(self,):
        self.meta = {}

    def transform(self, t):
        self.tree = t
        if not t.meta or not isinstance(t.meta, dict):
            raise Exception("needs meta")
        self.meta = t.meta
        # for k, v in t.meta["dependencies"].items():
        #     print(f"{k} -> {list(v)}")
        return super().transform(t)


def join_names(names):
    # print(repr(key))
    # print(self.dependencies[key])
    if not names:
        return str(uuid.uuid1())[:8]
    id = str(names[0]) + "".join([capitalize(x) for x in names[1:]])
    assert id
    return id


class Splitter(TransformerWithMeta):
    types: dict

    def __init__(
        self,
        objects_inside_objects=True,
        objects_inside_lists=True,
        unions_inside_objects=True,
        unions_inside_lists=True,
        join_names=join_names,
    ):
        self.types = {}
        self.join_names = join_names
        self.objects_inside_objects = objects_inside_objects
        self.objects_inside_lists = objects_inside_lists
        self.unions_inside_lists = unions_inside_lists
        self.unions_inside_objects = unions_inside_objects

    def make_new_name(self, name):
        if not isinstance(name, UniqueKey):
            name = UniqueKey(name)
        return self.join_names(list(self.dependencies[name]) + [name])

    @property
    def dependencies(self) -> dict:
        return self.meta["dependencies"]

    @v_args(tree=True)
    def root_pair(self, t):
        key, *_ = t.children
        self.types[key] = t
        return t

    def start(self, children):
        types = list(self.types.values())
        # tree.children.extend(types)
        return Tree("start", types)

    def object(self, children):
        if not self.objects_inside_objects:
            return Tree("object", children)
        for key in children:
            name, value = key.children  # TODO sometimes first is annotation
            if value.data in ["object"]:
                id = self.make_new_name(name)
                self.types[id] = Tree("root_pair", [id] + copy(key.children[1:],), meta={})
                key.children = [name, Tree("reference", [id])]
        return Tree("object", children)

    @v_args(meta=True)
    def list(self, children, meta):
        if not meta or isinstance(meta, Meta):
            logger.warn(f'list parent of {children[0].pretty()} has not meta')

        if not any([self.objects_inside_lists, self.unions_inside_lists]):
            return Tree(composed_types.LIST, children)
        TO_SPLIT = {
            composed_types.OBJECT: self.objects_inside_lists,
            composed_types.UNION: self.unions_inside_lists,
        }
        TO_SPLIT = [k for k, v in TO_SPLIT.items() if v]
        child, = children
        if not child.data in TO_SPLIT:
            return Tree("list", children)
        name = meta["parent_key"]
        id = self.make_new_name(name)
        old_child = copy(child)
        self.types[id] = Tree("root_pair", [id] + [old_child], meta={})
        return Tree("list", [Tree("reference", [id])])

    @v_args(tree=True)
    def union(self, t,):
        meta = t.meta
        children = t.children
        if not meta:
            logger.warning(f'no meta for {children} union')
            return t
        if not self.unions_inside_objects:
            return Tree("union", children)
        name = meta["parent_key"]
        id = self.make_new_name(name)
        old_children = copy(children)
        self.types[id] = Tree("root_pair", [id] + [Tree("union", old_children)], meta={})
        return Tree("reference", [id])



class modifiers:
    GRAPHQL_INPUT = "[graphql input]"
    GRAPHQL_HIDDEN = "[graphql hide]"


class types:
    STR = "type_str"
    BOOL = "type_bool"
    ANY = "type_any"
    INT = "type_int"
    FLOAT = "type_float"
    REGEX = "regex"


class literals:
    NULL = "literal_null"
    TRUE = "literal_true"
    FALSE = "literal_false"
    STRING = "literal_string"
    INTEGER = "literal_integer"
    ELLIPSIS = "literal_ellipsis"


class composed_types:
    OBJECT = "object"
    LIST = "list"
    INTERSECTION = "intersection"
    UNION = "union"
    RANGE = "bounded_range"
    LOW_RANGE = "low_bounded_range"
    HIGH_RANGE = "high_bounded_range"


class structure:
    START = "start"
    ROOT_PAIR = "root_pair"
    REQUIRED_PAIR = "required_pair"
    OPTIONAL_PAIR = "optional_pair"
    REFERENCE = "reference"
    ANNOTATION = "annotation"


def capitalize(s: str):
    if not s:
        return ""
    return s[0].capitalize() + s[1:]



from lark import Lark, Tree, Transformer, Token


class UniqueKey(Token):
    type: str
    value: str
    pos_in_stream: str
    line: str
    column: str
    end_line: str
    end_column: str

    def __new__(cls, token):
        self = Token.__new__(
            cls,
            token.type,
            token.value,
            token.pos_in_stream,
            token.line,
            token.column,
            token.end_line,
            token.end_column,
        )
        return self

    def __repr__(self):
        return f"`{self.value}`"

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __hash__(self):
        return str.__hash__(f"{self.column}{self.line}{self.value}")



import fire
import json
import sys
from skema.resolve_refs import resolve_refs
import skema.generators as gens
from skema.parser import parse


def as_list(x):
    if isinstance(x, (list, tuple)):
        return x
    return [x]


def write(path, data):
    with open(path) as f:
        f.write(data)


def get_stdin() -> str:
    if sys.stdin.isatty():
        print("need skema from stdin")
        return
    lines = list(sys.stdin)
    lines = [l for l in lines if l.strip()]
    string = "".join(lines) + "\n"
    return string


class Gen:

    def python(self,):
        stdin = get_stdin()
        tree = parse(stdin)
        code = gens.python(tree)
        print(code.strip())

    def typescript(self,):
        stdin = get_stdin()
        tree = parse(stdin)
        code = gens.typescript(tree)
        print(code.strip())

    def jsonschema(self, ref=None, resolve=False):
        stdin = get_stdin()
        tree = parse(stdin)
        obj = gens.jsonschema(tree, ref=ref, resolve=resolve)
        code = json.dumps(obj, indent=4)
        print(code.strip())

    def graphql(self,):
        stdin = get_stdin()
        tree = parse(stdin)
        code = gens.graphql(tree)
        print(code.strip())

class FakeData:
    pass

class InferJson:
    pass

class FromCode:
    pass

def print_tree():
    stdin = get_stdin()
    t = parse(stdin)
    print(t.pretty())

Cli = {
    'gen': Gen,
    'fakedata': FakeData, # cat schema.skema | skema fakedata json > data.json
    'tree': print_tree
    # 'from': FromCode, # skema from jsonschema -f schema.json -t skema.skema
    # 'inferjson': InferJson # cat data.json | skema inferjson > skema.skema
}


from skema.support import types, modifiers

import re
from itertools import chain
from typing import Any, Callable, Dict, List, Optional, Union, cast

from graphql.language import print_ast
from graphql.language.block_string import print_block_string
from graphql.pyutils import inspect
from graphql.type import (
    DEFAULT_DEPRECATION_REASON,
    GraphQLArgument,
    GraphQLDirective,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLInterfaceType,
    GraphQLNamedType,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLSchema,
    GraphQLString,
    GraphQLUnionType,
    is_enum_type,
    is_input_object_type,
    is_interface_type,
    is_introspection_type,
    is_object_type,
    is_scalar_type,
    is_specified_directive,
    is_specified_scalar_type,
    is_union_type,
)
from graphql.utilities.ast_from_value import ast_from_value

__all__ = ["print_schema", "print_introspection_schema", "print_type", "print_value"]


def print_schema(schema: GraphQLSchema) -> str:
    return print_filtered_schema(
        schema, lambda n: not is_specified_directive(n), is_defined_type
    )


def print_introspection_schema(schema: GraphQLSchema) -> str:
    return print_filtered_schema(schema, is_specified_directive, is_introspection_type)


def is_defined_type(type_: GraphQLNamedType) -> bool:
    return not is_specified_scalar_type(type_) and not is_introspection_type(type_)


def print_filtered_schema(
    schema: GraphQLSchema,
    directive_filter: Callable[[GraphQLDirective], bool],
    type_filter: Callable[[GraphQLNamedType], bool],
) -> str:
    directives = filter(directive_filter, schema.directives)
    type_map = schema.type_map
    types = filter(type_filter, map(type_map.get, sorted(type_map)))  # type: ignore

    return (
        "\n\n".join(
            chain(
                filter(None, [print_schema_definition(schema)]),
                (print_directive(directive) for directive in directives),
                (print_type(type_) for type_ in types),  # type: ignore
            )
        )
        + "\n"
    )


def print_schema_definition(schema: GraphQLSchema) -> Optional[str]:
    if is_schema_of_common_names(schema):
        return None

    operation_types = []

    query_type = schema.query_type
    if query_type:
        operation_types.append(f"  query: {query_type.name}")

    mutation_type = schema.mutation_type
    if mutation_type:
        operation_types.append(f"  mutation: {mutation_type.name}")

    subscription_type = schema.subscription_type
    if subscription_type:
        operation_types.append(f"  subscription: {subscription_type.name}")

    return "schema {\n" + "\n".join(operation_types) + "\n}"


def is_schema_of_common_names(schema: GraphQLSchema) -> bool:
    """Check whether this schema uses the common naming convention.

    GraphQL schema define root types for each type of operation. These types are the
    same as any other type and can be named in any manner, however there is a common
    naming convention:

    schema {
      query: Query
      mutation: Mutation
    }

    When using this naming convention, the schema description can be omitted.
    """
    query_type = schema.query_type
    if query_type and query_type.name != "Query":
        return False

    mutation_type = schema.mutation_type
    if mutation_type and mutation_type.name != "Mutation":
        return False

    subscription_type = schema.subscription_type
    if subscription_type and subscription_type.name != "Subscription":
        return False

    return True


def print_type(type_: GraphQLNamedType) -> str:
    if is_scalar_type(type_):
        type_ = cast(GraphQLScalarType, type_)
        return print_scalar(type_)
    if is_object_type(type_):
        type_ = cast(GraphQLObjectType, type_)
        return print_object(type_)
    if is_interface_type(type_):
        type_ = cast(GraphQLInterfaceType, type_)
        return print_interface(type_)
    if is_union_type(type_):
        type_ = cast(GraphQLUnionType, type_)
        return print_union(type_)
    if is_enum_type(type_):
        type_ = cast(GraphQLEnumType, type_)
        return print_enum(type_)
    if is_input_object_type(type_):
        type_ = cast(GraphQLInputObjectType, type_)
        return print_input_object(type_)
    # Not reachable. All possible types have been considered.
    raise TypeError(f"Unexpected type: '{inspect(type_)}'.")  # pragma: no cover


def print_scalar(type_: GraphQLScalarType) -> str:
    return print_description(type_) + f"{type_.name}: Any"


def print_object(type_: GraphQLObjectType) -> str:
    interfaces = type_.interfaces
    implemented_interfaces = (
        (" &".join(f' {i.name}' for i in interfaces)  + ' &') if interfaces else ""
    )
    return (
        print_description(type_)
        + f"{type_.name}:" +  implemented_interfaces
        + print_fields(type_)
    )


def print_interface(type_: GraphQLInterfaceType) -> str:
    return print_description(type_) + f"{type_.name}:" + print_fields(type_)


def print_union(type_: GraphQLUnionType) -> str:
    types = type_.types
    possible_types = ": " + " | ".join(t.name for t in types) if types else ""
    return print_description(type_) + f"{type_.name}" + possible_types


def print_enum(type_: GraphQLEnumType) -> str:
    values = [
        print_description(value, "  ", not i) + f"  {name}" + print_deprecated(value)
        for i, (name, value) in enumerate(type_.values.items())
    ]
    return print_description(type_) + f"{type_.name}: " + ' | '.join([f'"{x}"' for x in values])


def print_input_object(type_: GraphQLInputObjectType) -> str:
    fields = [
        print_description(field, "    ", not i) + "    " + print_input_value(name, field)
        for i, (name, field) in enumerate(type_.fields.items())
    ]
    return print_description(type_, adding='[graphql input]') + f"{type_.name}:" + print_block(fields)


map_graphql_to_skema = {
    'String': 'Str',
    'Boolean': 'Bool',
    'Json': 'Any',
    'Int': 'Int',
    'Float': 'Float',
    'ID': 'Str',
}

def print_fields(type_: Union[GraphQLObjectType, GraphQLInterfaceType]) -> str:
    fields = [
        print_description(field, "  ", not i)
        + f"    {name}"
        + print_args(field.args, "  ")
        + f": {map_graphql_to_skema.get(str(field.type), field.type)}"
        + print_deprecated(field)
        for i, (name, field) in enumerate(type_.fields.items())
    ]
    return print_block(fields)


def print_block(items: List[str]) -> str:
    return "\n" + "\n".join(items) + "\n" if items else ""


def print_args(args: Dict[str, GraphQLArgument], indentation="") -> str:
    if not args:
        return ""

    # If every arg does not have a description, print them on one line.
    if not any(arg.description for arg in args.values()):
        return (
            "("
            + ", ".join(print_input_value(name, arg) for name, arg in args.items())
            + ")"
        )

    return (
        "(\n"
        + "\n".join(
            print_description(arg, f"  {indentation}", not i)
            + f"  {indentation}"
            + print_input_value(name, arg)
            for i, (name, arg) in enumerate(args.items())
        )
        + f"\n{indentation})"
    )


def print_input_value(name: str, arg: GraphQLArgument) -> str:
    default_ast = ast_from_value(arg.default_value, arg.type)
    arg_decl = f"{name}: {arg.type}"
    if default_ast:
        arg_decl += f" = {print_ast(default_ast)}"
    return arg_decl


def print_directive(directive: GraphQLDirective) -> str:
    return ''
    # return (
    #     print_description(directive)
    #     + f"directive @{directive.name}"
    #     + print_args(directive.args)
    #     + (" repeatable" if directive.is_repeatable else "")
    #     + " on "
    #     + " | ".join(location.name for location in directive.locations)
    # )


def print_deprecated(field_or_enum_value: Union[GraphQLField, GraphQLEnumValue]) -> str:
    return ''
    # if not field_or_enum_value.is_deprecated:
    #     return ""
    # reason = field_or_enum_value.deprecation_reason
    # reason_ast = ast_from_value(reason, GraphQLString)
    # if not reason_ast or reason == "" or reason == DEFAULT_DEPRECATION_REASON:
    #     return " @deprecated"
    # return f" @deprecated(reason: {print_ast(reason_ast)})"


def print_description(
    def_: Union[GraphQLArgument, GraphQLDirective, GraphQLEnumValue, GraphQLNamedType],
    indentation="",
    first_in_block=True,
    adding='',
) -> str:
    if not def_.description:
        return ""

    lines = description_lines(def_.description, 120 - len(indentation))
    if adding:
        lines += ['\n' + adding + '\n']
    text = "\n".join(lines)
    prefer_multiple_lines = len(text) > 70
    block_string = print_block_string(text, "", prefer_multiple_lines)
    prefix = "\n" + indentation if indentation and not first_in_block else indentation

    return prefix + block_string.replace("\n", "\n" + indentation) + "\n"


def escape_quote(line: str) -> str:
    return line.replace('"""', '\\"""')


def description_lines(description: str, max_len: int) -> List[str]:
    lines: List[str] = []
    append_line, extend_lines = lines.append, lines.extend
    raw_lines = description.splitlines()
    for raw_line in raw_lines:
        if raw_line:
            # For > 120 character long lines, cut at space boundaries into sublines
            # of ~80 chars.
            extend_lines(break_line(raw_line, max_len))
        else:
            append_line(raw_line)
    return lines


def break_line(line: str, max_len: int) -> List[str]:
    if len(line) < max_len + 5:
        return [line]
    parts = re.split(f"((?: |^).{{15,{max_len - 40}}}(?= |$))", line)
    if len(parts) < 4:
        return [line]
    sublines = [parts[0] + parts[1] + parts[2]]
    append_subline = sublines.append
    for i in range(3, len(parts), 2):
        append_subline(parts[i][1:] + parts[i + 1])
    return sublines


def print_value(value: Any, type_: GraphQLInputType) -> str:
    """Convenience function for printing a Python value"""
    return print_ast(ast_from_value(value, type_))  # type: ignore



from lark import Visitor, Tree, Token, v_args
from .support import Transformer
from functools import partial
from funcy import cat, flip, collecting
from prtty import pretty
from collections import defaultdict
from toposort import toposort, toposort_flatten
from ordered_set import OrderedSet
from ..types import UniqueKey
import uuid
from copy import copy
from ..support import structure, composed_types


class SortOptionalsLast(Transformer):
    def object(self, children):
        optionals = [
            child for child in children if child.data == structure.OPTIONAL_PAIR
        ]
        required = [
            child for child in children if child.data == structure.REQUIRED_PAIR
        ]
        return Tree(composed_types.OBJECT, required + optionals)



from .support import Transformer
from .merge_intersections import MergeIntersections
from .metas import AddListMetas, GetDependencies, AddUnionMetas, RemoveAnnotations, RemoveEllipses
from .splitter import Splitter
from .support import Printer
from .sort import SortOptionalsLast



from skema.reconstruct import from_jsonschema
from graphql import build_schema
from ..support import *


def test_1():
    print()
    t = from_jsonschema(o2)
    print(t.pretty())
    print(gens.typescript(t))


false = False
true = True
null = None
o1 = {
    "title": "Object",
    "description": "",
    "type": "object",
    "properties": {
        "xxx": {"type": "string", "description": ""},
        "yyy": {"description": ""},
    },
    "additionalProperties": false,
}

o2 = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$ref": "#/definitions/SchemaTypeEnums",
    "definitions": {
        "SchemaTypeEnums": {
            "title": "SchemaTypeEnums",
            "description": "",
            "enum": [
                {"const": "string"},
                {"const": "array"},
                {"const": "object"},
                {"const": "number"},
                {"const": "boolean"},
                {"const": "integer"},
            ],
        },
        "SchemaType": {
            "title": "SchemaType",
            "description": "",
            "anyOf": [
                {"$ref": "#/definitions/SchemaTypeEnums"},
                {"type": "array", "items": {"$ref": "#/definitions/SchemaTypeEnums"}},
            ],
        },
        "Block": {
            "title": "Block",
            "description": "",
            "anyOf": [
                {"$ref": "#/definitions/GenericBlock"},
                {"$ref": "#/definitions/Enum"},
                {"$ref": "#/definitions/Object"},
                {"$ref": "#/definitions/Array"},
                {"$ref": "#/definitions/Constant"},
                {"$ref": "#/definitions/AllOf"},
                {"$ref": "#/definitions/AnyOf"},
                {"$ref": "#/definitions/OneOf"},
            ],
        },
        "Object": {
            "title": "Object",
            "description": "",
            "type": "object",
            "properties": {
                "type": {"$ref": "#/definitions/SchemaType", "description": ""},
                "properties": {"description": ""},
                "additionalProperties": {"description": ""},
            },
            "additionalProperties": false,
            "required": ["type", "properties", "additionalProperties"],
        },
        "GenericBlock": {
            "title": "GenericBlock",
            "description": "",
            "type": "object",
            "properties": {
                "type": {"$ref": "#/definitions/SchemaType", "description": ""}
            },
            "additionalProperties": false,
            "required": ["type"],
        },
        "Enum": {
            "title": "Enum",
            "description": "",
            "type": "object",
            "properties": {
                "enum": {
                    "type": "array",
                    "items": {"$ref": "#/definitions/Block"},
                    "description": "",
                }
            },
            "additionalProperties": false,
            "required": ["enum"],
        },
        "Constant": {
            "title": "Constant",
            "description": "",
            "type": "object",
            "properties": {"const": {"type": "string", "description": ""}},
            "additionalProperties": false,
            "required": ["const"],
        },
        "AllOf": {
            "title": "AllOf",
            "description": "",
            "type": "object",
            "properties": {
                "allOf": {
                    "type": "array",
                    "items": {"$ref": "#/definitions/Block"},
                    "description": "",
                }
            },
            "additionalProperties": false,
            "required": ["allOf"],
        },
        "AnyOf": {
            "title": "AnyOf",
            "description": "",
            "type": "object",
            "properties": {
                "anyOf": {
                    "type": "array",
                    "items": {"$ref": "#/definitions/Block"},
                    "description": "",
                }
            },
            "additionalProperties": false,
            "required": ["anyOf"],
        },
        "OneOf": {
            "title": "OneOf",
            "description": "",
            "type": "object",
            "properties": {
                "oneOf": {
                    "type": "array",
                    "items": {"$ref": "#/definitions/Block"},
                    "description": "",
                }
            },
            "additionalProperties": false,
            "required": ["oneOf"],
        },
        "Array": {
            "title": "Array",
            "description": "",
            "type": "object",
            "properties": {"items": {"$ref": "#/definitions/Block", "description": ""}},
            "additionalProperties": false,
            "required": ["items"],
        },
    },
}



from .parser import parse
from lark.visitors import TransformerChain
import skema.languages as l
import skema.transformers as t
from .resolve_refs import resolve_refs


def jsonschema(tree, ref=None, resolve=False):
    transformer = TransformerChain(
        t.RemoveAnnotations(), t.RemoveEllipses(), l.JsonSchema(ref=ref)
    )
    data = transformer.transform(tree)
    if resolve:
        resolve_refs(data)
    return data


def python(tree):
    transformer = TransformerChain(
        t.RemoveAnnotations(),
        t.RemoveEllipses(),
        t.MergeIntersections(),
        t.GetDependencies(),
        t.AddListMetas(),
        t.AddUnionMetas(),
        t.Splitter(unions_inside_objects=False, unions_inside_lists=False),
        t.SortOptionalsLast(),
        l.AddInitializersMetas(),
        l.Python(),
    )

    return transformer.transform(tree)


def typescript(tree):
    transformer = TransformerChain(
        t.RemoveAnnotations(),
        t.RemoveEllipses(),
        t.MergeIntersections(),
        t.GetDependencies(),
        t.AddListMetas(),
        t.AddUnionMetas(),
        t.Splitter(unions_inside_objects=False),
        l.Typescript(),
    )
    return transformer.transform(tree)


def graphql(tree):
    transformer = TransformerChain(
        t.RemoveAnnotations(),
        t.RemoveEllipses(),
        t.MergeIntersections(),
        t.GetDependencies(),
        t.AddListMetas(),
        t.AddUnionMetas(),
        t.Splitter(),
        t.Printer(),
        l.Graphql(),
    )

    return transformer.transform(tree)



from prtty import pretty
from populate import indent_to, populate_string
from lark import Token, Tree, v_args
from funcy import merge, lmap, omit, concat
from ..transformers import Transformer
from ..parser import parser

ELLIPSIS = "..."


class Typescript(Transformer):
    def __init__(self, ref=None):
        self.ref = ref

    def __default__(self, data, children, meta):
        raise NotImplementedError(f"{data} is not implemented in grahql") from None

    def start(self, children):
        return "\n".join(children)

    def type_str(self, _):
        return "string"

    def type_bool(self, _):
        return "boolean"

    def type_int(self, _):
        return "number"

    def type_float(self, _):
        return "number"

    def type_any(self, _):
        return "any"

    def literal_null(self, _):
        return "null"

    def literal_true(self, _):
        return f"true"

    def literal_false(self, _):
        return f"false"

    def literal_string(self, children):
        value, = children
        return f'{value}'

    def literal_number(self, children):
        value, = children
        return str(value)

    def literal_ellipsis(self, _):
        return ELLIPSIS

    def regex(self, children):
        # value, = children
        return "string"

    def reference(self, children):
        value, = children
        return value

    def annotation(self, children):
        value, = children
        return value

    # TODO remove and make range better
    def scalar(self, children):
        value, = children
        return value

    def object(self, children):
        types = "\n    ".join(children)  # .replace(' ', '.')
        return "export interface $key {\n    " + types + "\n}\n"

    def list(self, children):
        value, = children
        return f"Array<{value}>"

    def union(self, children):
        return " | ".join(children)

    def required_pair(self, children):
        k, v = children
        if "$key" in v:
            return v.replace("$key", k)
        return k + ": " + v

    def optional_pair(self, children):
        k, v = children
        if "$key" in v:
            return v.replace("$key", k)
        return k + "?: " + v

    @v_args(meta=True)
    def root_pair(self, children, meta):
        k, v = children
        annotation = meta['annotation'] if 'annotation' in meta else ''    
        annotation = annotation and f'/*\n{annotation}\n*/\n'
        if "$key" in v:
            return annotation + v.replace("$key", k)
        else:
            return annotation + f"export type {k} = {v}\n"

    pass


def get_first_key(obj):
    keys = list(obj.keys())
    return keys[0]



from skema.parser import parse
from random import randint
from funcy import omit
from lark import Tree, Token
import skema.reconstruct.schema_types as jsontypes
from ..support import composed_types, types, literals, structure
from ..dictlike import dictlike


def is_block(obj):
    ks = ["anyOf", "allOf", "oneOf", "type", "const"]
    return any([k in obj for k in ks])


def reconstruct(obj, definition_key="definitions", root_name="Root") -> Tree:
    nodes = []
    obj = dictlike(**obj)
    if definition_key in obj:
        for name, block in obj[definition_key].items():
            block = dictlike(block)
            t = Tree(structure.ROOT_PAIR, [make_token(name), process_block(block)])
            nodes.append(t)
    if is_block(obj):
        if "title" in obj:
            root_name = obj['title']
        obj = omit(obj, ["$ref"])
        obj = dictlike(obj)
        t = Tree(structure.ROOT_PAIR, [make_token(root_name), process_block(obj)])
        nodes.append(t)
    return Tree(structure.START, nodes)
    # print(skema.replace(' ', '.'))

    # return parse(skema)


def get_ref_name(block):
    return block["$ref"].split("/")[-1]


def process_block(block: jsontypes.Block,):
    blocktype = block.get("type", "")
    if False:
        pass
    elif "$ref" in block:
        name = get_ref_name(block)
        return Tree(structure.REFERENCE, [name])
    elif "anyOf" in block or "oneOf" in block:
        k = 'anyOf' if block.get("anyOf") else 'oneOf'
        return Tree(
            composed_types.UNION, [process_block(dictlike(item)) for item in block[k]]
        )
    elif "allOf" in block:
        return Tree(
            composed_types.INTERSECTION,
            [process_block(dictlike(item)) for item in block.anyOf],
        )
    elif "enum" in block:
        return Tree(composed_types.UNION, [process_block(dictlike(x)) for x in block.enum])
    elif "const" in block:
        x = block.const
        data = literals.INTEGER 
        if isinstance(x, (float,)):
            raise Exception('const cannot contina floats')
        if isinstance(x, (int,)):
            data = literals.INTEGER
        if isinstance(x, (str,)):
            data = literals.STRING
            x = f'"{x}"'
        return Tree(data, [x])
    elif blocktype == "array":
        return Tree(composed_types.LIST, [process_block(dictlike(block["items"]))])
    elif blocktype == "object":
        pairtype = structure.REQUIRED_PAIR
        properties = block.get("properties", {})
        return Tree(
            composed_types.OBJECT,
            [
                Tree(pairtype, [make_token(k), process_block(dictlike(v))])
                for k, v in properties.items()
            ],
        )
    elif blocktype == "string":
        return make_type_tree(types.STR)
    elif blocktype == "integer":
        return make_type_tree(types.INT)
    elif blocktype == "number":
        return make_type_tree(types.FLOAT)
    elif blocktype == "boolean":
        return make_type_tree(types.BOOL)
    else:
        return make_type_tree(types.ANY)


def make_token(data):
    return Token("UNKNOWN", data, line=randint(1, 999), column=randint(1, 999))


def make_type_tree(data):
    data = make_token(data)
    return Tree(data, [])


"""
type == integer -> 
type == number -> 
type == string -> 
type == array -> Tree(list, [process_block(obj.items)])
type == object -> Tree(object, [Tree(pair, k, [process_block(v)] for k, v in obj.properties.items()]) 
type == or -> [Tree(union, [process_block(item)]) for item in obj.anyOf]
"""



from skema import dictlike
from typing import Optional, List, Any, Union
from typing_extensions import Literal

SchemaTypeEnums = Union[Literal["string"], Literal["array"], Literal["object"], Literal["number"], Literal["boolean"], Literal["integer"]]
SchemaTypeEnums.from_ = lambda x: x

SchemaType = Union["SchemaTypeEnums", List["SchemaTypeEnums"]]
SchemaType.from_ = lambda x: x

Block = Union["GenericBlock", "Enum", "Object", "Array", "Constant", "AllOf", "AnyOf", "OneOf"]
Block.from_ = lambda x: x

class Object(dictlike):
    
    type: "SchemaType"
    properties: Any
    additionalProperties: Any
    
    def __init__(
        self,
        *, 
        type,
        properties,
        additionalProperties, 
        **kwargs
    ):
        super().__init__(
            type=SchemaType.from_(type) if type != None else None,
            properties=properties if properties != None else None,
            additionalProperties=additionalProperties if additionalProperties != None else None,
            **kwargs
        )

class GenericBlock(dictlike):
    
    type: "SchemaType"
    
    def __init__(
        self,
        *, 
        type, 
        **kwargs
    ):
        super().__init__(
            type=SchemaType.from_(type) if type != None else None,
            **kwargs
        )

class Enum(dictlike):
    
    enum: List["Block"]
    
    def __init__(
        self,
        *, 
        enum, 
        **kwargs
    ):
        super().__init__(
            enum=[Block.from_(x) for x in enum] if enum != None else None,
            **kwargs
        )

class Constant(dictlike):
    
    const: str
    
    def __init__(
        self,
        *, 
        const, 
        **kwargs
    ):
        super().__init__(
            const=str(const) if const != None else None,
            **kwargs
        )

class AllOf(dictlike):
    
    allOf: List["Block"]
    
    def __init__(
        self,
        *, 
        allOf, 
        **kwargs
    ):
        super().__init__(
            allOf=[Block.from_(x) for x in allOf] if allOf != None else None,
            **kwargs
        )

class AnyOf(dictlike):
    
    anyOf: List["Block"]
    
    def __init__(
        self,
        *, 
        anyOf, 
        **kwargs
    ):
        super().__init__(
            anyOf=[Block.from_(x) for x in anyOf] if anyOf != None else None,
            **kwargs
        )

class OneOf(dictlike):
    
    oneOf: List["Block"]
    
    def __init__(
        self,
        *, 
        oneOf, 
        **kwargs
    ):
        super().__init__(
            oneOf=[Block.from_(x) for x in oneOf] if oneOf != None else None,
            **kwargs
        )

class Array(dictlike):
    
    items: "Block"
    
    def __init__(
        self,
        *, 
        items, 
        **kwargs
    ):
        super().__init__(
            items=Block.from_(items) if items != None else None,
            **kwargs
        )



# from lark import Visitor, Tree, Transformer, Token, v_args
# from functools import partial
# from funcy import cat, flip, collecting
# from prtty import pretty
# from collections import defaultdict
# from toposort import toposort, toposort_flatten
# from ordered_set import OrderedSet
# from .types import UniqueKey
# import uuid
# from copy import copy


# # def make_reference(id, tree: Tree):
# #     old_children = copy(tree.children)
# #     tree.children = [Tree("reference", [id])]
# #     subs = Tree("root_pair", [id] + old_children)
# #     return subs


# # class MakeMap(Visitor):
# #     types: dict

# #     def __init__(self):
# #         self.types = {}
# #         pass

# #     def root_pair(self, tree):
# #         key, *_ = tree.children
# #         self.types[key] = tree


# # i could remove this class by
# # getting the path from root to leaf for the node i am removing,
# # make the name of the reference as the join of the parent names
# class ReplaceIds(Transformer):
#     types: dict = {}
#     child_of: defaultdict = defaultdict(set)

#     def __init__(self,):
#         pass

#     def start(self, children):
#         translations = {}
#         ids_items = [
#             (k, {x for x in v if x in self.child_of.keys()})
#             for k, v in self.child_of.items()
#         ]
#         ids_sorted = topological_sort(ids_items)
#         ids_sorted = list(ids_sorted)
#         for id in ids_sorted:
#             translations[id] = "_".join(reversed(list(self.child_of[id])))
#             self.types[id].children[0] = translations[id]

#             for _, parents in self.child_of.items():
#                 if id in parents:
#                     parents.remove(id)
#                     parents.add(translations[id])

#             for parent in self.child_of[id]:
#                 original_id = flip(translations).get(parent)
#                 if not original_id in self.types:
#                     continue
#                 refs = self.types[original_id].find_pred(
#                     lambda node: node.data == "reference"
#                 )
#                 for ref_node in refs:
#                     refname, = ref_node.children
#                     if refname in translations:
#                         ref_node.children = [translations[refname]]
#         return Tree("start", list(self.types.values()))

#     def root_pair(self, children):
#         key, *_ = children
#         self.types[key] = Tree("root_pair", children)

#         # get generic references
#         refs_parents = Tree("", children).find_pred(is_reference_parent)
#         for child in refs_parents:
#             parentname, ref = child.children  #  TODO first child might be annotation
#             refname, = ref.children
#             self.child_of[refname].add(str(key))
#             self.child_of[refname].add(str(parentname))

#         # get list references
#         list_refs_parents = list(Tree("", children).find_pred(is_reference_list_parent))
#         for list_parent in list_refs_parents:
#             parentname, list_node, = list_parent.children
#             ref, = list_node.children
#             refname, = ref.children
#             self.child_of[refname].add(str(key))
#             self.child_of[refname].add(str(parentname))

#         # if refs_parents:
#         #     pretty(self.child_of)
#         #     pretty(list(self.types.keys()))
#         return Tree("root_pair", children)


# # class SplitterVisitor(Visitor):
# #     types: dict = {}


# #     def root_pair(self, tree):
# #         key, *_ = tree.children
# #         self.types[key] = tree

# #     def start(self, tree: Tree):
# #         types = list(self.types.values())
# #         # tree.children.extend(types)
# #         tree.children = self.types.values()

# #     def __init__(self, ):
# #         self.types = {}
# #         pass

# #     def list(self, tree: Tree):
# #         child, = tree.children
# #         id = make_id()
# #         if child.data in ["object", "list"]:
# #             self.types[id] = make_reference(id, tree)
# #         pass

# #     def intersection(self, tree: Tree):
# #         raise Exception('cannot handle intersections')

# #     def object(self, tree: Tree):
# #         for key in tree.children:
# #             name, value = key.children # TODO sometimes first is annotation
# #             if value.data in ['object']:
# #                 print('ok')
# #                 id = make_id()
# #                 self.types[id] = Tree("root_pair", [id] + copy(key.children[1:]))
# #                 key.children = [name, Tree("reference", [id])]



from lark import Visitor, Tree, Token, v_args
from .support import Transformer, unique
from functools import partial
from funcy import cat, flip, collecting
from prtty import pretty
from collections import defaultdict
from toposort import toposort, toposort_flatten
from ordered_set import OrderedSet
from ..types import UniqueKey
import uuid
from copy import copy


class MergeIntersections(Transformer):
    types: dict

    def __init__(self,):
        self.types = {}

    @v_args(tree=True)
    def root_pair(self, t):
        key, *_ = t.children
        self.types[key] = t
        # is_reference = lambda node: node.data == "reference"
        # for child in t.find_pred(is_reference):
        #     print(child)
        #  self.child_of
        return t

    # def intersection(self, tree: Tree):
    #     to_join = []
    #     for child in tree.children:
    #         if child.data == 'reference':
    #             ref, = child.children
    #             root_pair = self.types[ref]
    #             # only if object
    #             _, object = root_pair.children
    #             fields = object.children
    #             to_join += fields
    #         elif child.data == 'object':
    #             to_join += child.children
    #     tree.data = 'object'
    #     tree.children = to_join

    def intersection(self, children):
        to_join = []
        for child in children:
            if child.data == "reference":
                ref, = child.children
                root_pair = self.types[ref]
                # only if object
                _, object = root_pair.children
                fields = object.children
                to_join += fields
            elif child.data == "object":
                to_join += child.children
        children = unique(to_join, key=lambda x: x.children[0])
        return Tree("object", children)



from prtty import pretty
from lark import Token, Tree, v_args
from ..transformers import Transformer
from funcy import merge, lmap, omit, concat
from ..parser import parser
from ..support import modifiers

ELLIPSIS = "..."


class Graphql(Transformer):
    def __init__(self, ref=None):
        self.ref = ref

    def __default__(self, data, children, meta):
        raise NotImplementedError(f"{data} is not implemented in grahql") from None

    def start(self, children):
        return "\n".join(children)

    def type_str(self, _):
        return "String"

    def type_bool(self, _):
        return "Boolean"

    def type_int(self, _):
        return "Int"

    def type_float(self, _):
        return "Float"

    def type_any(self, _):
        return "Json"

    # def literal_null(self, _):
    #     raise NotImplementedError("null not exists in graphql")

    # def literal_true(self, _):
    #     raise NotImplementedError("true not exists in graphql")

    # def literal_false(self, _):
    #     raise NotImplementedError("false not exists in graphql")
    @v_args(tree=True)
    def literal_string(self, t):
        v, = t.children
        return v

    def literal_ellipsis(self, _):
        return ELLIPSIS

    def regex(self, children):
        # value, = children
        return "String"

    def reference(self, children):
        value, = children
        return str(value)

    def annotation(self, children):
        value, = children
        return str(value)

    # TODO remove and make range better
    def scalar(self, children):
        value, = children
        return value

    def object(self, children):
        return "type $key {\n" + "\n".join(["    " + c for c in children]) + "\n}\n"

    def list(self, children):
        value, = children
        return f"[{value}]"

    def union(self, children):
        if all([isinstance(x, str) and '"' in x for x in children]):
            return (
                "enum $key {\n"
                + "\n".join(["    " + str(x[1:-1]) for x in children])
                + "\n}\n"
            )
        elif all([isinstance(x, str) for x in children]):
            return "union $key = " + " | ".join(children) + "\n"
        else:
            raise NotImplementedError("cannot mix literals and shit")

    def required_pair(self, children):
        k, v = children
        if "$key" in v:
            return v.replace("$key", k)
        if '"' in v:
            v = 'String'
        return k + ": " + v

    optional_pair = required_pair

    @v_args(meta=True)
    def root_pair(self, children, meta):
        k, v = children
        annotation = meta['annotation'] if 'annotation' in meta else ''    
        annotation = annotation and f'"""\n{annotation}\n"""\n'
        if modifiers.GRAPHQL_HIDDEN in annotation: # TODO graphql hide should be made before splitting
            return ''

        if modifiers.GRAPHQL_INPUT in annotation:
            v = v.replace('type ', 'input ')
            
        if "$key" in v:
            return annotation + v.replace("$key", k)
        else:
            return annotation + f"scalar {k}\n"

    pass


def get_first_key(obj):
    keys = list(obj.keys())
    return keys[0]



import skema.generators as gens
from skema.parser import parse
from jsonschema import validate
import json
from typing import List
from hypothesis import given, settings
from random import randint
from funcy import omit
from hypothesis import strategies as hs
import json
from .regex import regex


def gen_int(prop):
    min_value = prop.get("minimum", 0)
    max_value = (
        prop.get("maximum")
        or prop.get("exclusiveMaximum")
        or (20 if not min_value else None)
    )
    return hs.integers(min_value=min_value, max_value=max_value)


def gen_string(prop):
    min_value = prop.get("minLength", 5)
    max_value = prop.get("maxLength", 20)
    # pattern = prop.get('pattern', r'[ A-Za-z\d]+')
    # return regex(pattern, )\
    #         .filter(lambda x: min_value <= len(x) and len(x) <= max_value)\
    #         .map(lambda x: x.rstrip('\x00'))
    return (
        hs.text(
            hs.characters(min_codepoint=97, max_codepoint=122),
            min_size=min_value,
            max_size=max_value
            # blacklist_categories=('Cc', 'Cs')),
        )
        .map(lambda s: s.strip())
        .filter(lambda s: len(s) > min_value)
    )


def should_include(key, required_list):
    if key in required_list:
        return True
    else:
        return bool(randint(0, 1))


def gen_array(prop, customs):
    min_items = prop.get("minItems", 2)
    max_items = prop.get("maxItems", 5)
    if prop.get("items", {}):
        generator = get_generator(prop.get("items"), customs)
        return hs.lists(elements=generator, min_size=min_items, max_size=max_items)
    else:
        return hs.lists(elements=gen_anything(), min_size=min_items, max_size=max_items)


def gen_anything():
    return hs.one_of(gen_int({}), gen_string({}), hs.booleans(), hs.none())


def gen_json_values():
    return (
        gen_string({})
        | hs.booleans()
        | gen_int({})
        | hs.none()
        | hs.floats(allow_nan=False, allow_infinity=False)
    )


def gen_any_obj():
    return hs.recursive(
        hs.dictionaries(gen_string({}), gen_json_values()),
        lambda children: hs.dictionaries(gen_string({}), children),
        max_leaves=10,
    )


def gen_object(prop, customs):
    required = prop.get("required", [])
    output = {}
    prop_key = "properties"
    if prop.get("properties", None) is None:
        if prop["additionalProperties"]:
            return gen_any_obj()

    for k in prop[prop_key].keys():
        json_prop = prop[prop_key][k]

        if should_include(k, required):
            output[k] = get_generator(json_prop, customs)

    return hs.fixed_dictionaries(output)


def gen_enum(prop):
    enum = prop["enum"]
    return hs.sampled_from(enum)


def gen_const(prop):
    const = prop["const"]
    return hs.just(const)


def gen_bool(prop):
    return hs.booleans()


def gen_one_of(prop, customs):
    possible_values = []
    for value in prop["oneOf"]:
        possible_values.append(get_generator(value, customs))

    return hs.one_of(possible_values)


def gen_any_of(prop, customs):
    possible_values = []
    for value in prop["anyOf"]:
        possible_values.append(get_generator(value, customs))

    return hs.one_of(possible_values)


specials = ["allOf", "oneOf", "anyOf"]


def merge_all_of(prop,):
    dicts = prop["allOf"]

    if not all(
        [any([x in d for x in specials]) or d.get("type") == "object" for d in dicts]
    ):
        raise Exception(
            f"allOf is supported as array of object types: {json.dumps(dicts, indent=4, default=repr)}"
        )
    result = {"additionalProperties": True, "required": [], "properties": {}}
    for obj in dicts:
        if "allOf" in obj:
            obj = merge_all_of(obj)

        result["additionalProperties"] = result["additionalProperties"] and bool(
            obj.get("additionalProperties", True)
        )
        result["properties"] = {
            **result["properties"],
            **(obj.get("properties", {}) or {}),
        }
        result["required"] = result["required"] + (obj.get("required", []) or [])
    result["required"] = list(set(result["required"]))
    return result


def gen_all_of(prop, customs):
    result = merge_all_of(prop)
    return gen_object(result, customs)


def get_generator(prop, customs={}):
    disp = {
        "string": gen_string,
        "integer": gen_int,
        "number": gen_int,
        "boolean": gen_bool,
    }
    if prop.get("title") != None:
        title = prop.get("title", "")  #  .strip()
        if title in customs:
            return hs.builds(lambda x: customs[title](), hs.integers())
    if not omit(prop, ["title", "description"]):
        return gen_anything()

    enum = prop.get("enum", None)
    if enum is not None:
        return gen_enum(prop)

    if "const" in prop:
        return gen_const(prop)

    one_of = prop.get("oneOf", None)
    if one_of is not None:
        return gen_one_of(prop, customs)

    any_of = prop.get("anyOf", None)
    if any_of is not None:
        return gen_any_of(prop, customs)

    all_of = prop.get("allOf", None)
    if all_of is not None:
        return gen_all_of(prop, customs)

    if "object" == prop.get("type"):
        return gen_object(prop, customs)

    if "array" == prop.get("type"):
        return gen_array(prop, customs)

    json_type = prop.get("type", None)
    if json_type is None:
        raise Exception("Couldnt find type in prop {0}".format(prop))

    return disp[json_type](prop)


def fake_data(
    schema: str, ref=None, amount=5, resolvers={}, from_json=False
) -> List[dict]:
    if from_json:
        json_schema = schema
    else:
        json_schema = gens.jsonschema(parse(schema), ref=ref, resolve=True)
    examples = []

    @given(get_generator(json_schema, resolvers))
    @settings(max_examples=amount)
    def generate(example_data):
        examples.append(example_data)

    generate()
    return examples




from lark import Visitor, Tree, Transformer as _Transformer, Token, v_args
from functools import partial
from funcy import cat, flip, collecting
from prtty import pretty
from collections import defaultdict
from toposort import toposort, toposort_flatten
from ordered_set import OrderedSet
from ..types import UniqueKey
from ..logger import logger
from ..support import structure
import uuid
from copy import copy


class Transformer(_Transformer):
    def __default__(self, data, children, meta):
        "Default operation on tree (for override)"
        if not isinstance(meta, dict):
            meta = {}
        return Tree(data, children, meta)


@v_args(tree=True)
class Printer(Transformer):
    def start(self, t):
        logger.debug(t.pretty())
        return t
    


@collecting
def unique(l, *, key):
    passed = set()
    for e in l:
        d = key(e)
        if not d in passed:
            yield e
        passed.add(d)


is_reference_parent = lambda node: (  # TODO add list reference k case
    node.data in [structure.REQUIRED_PAIR, structure.OPTIONAL_PAIR]
    and node.children[1].data == structure.REFERENCE
)

# is_reference_list_parent = lambda node: ( # TODO add list reference k case
#     node.data in ['optional_pair', 'required_pair'] and node.children[1].data == 'list'
#     and (node.children[1].children[0].data == 'list')
# )


def is_reference_list_parent(node: Tree):
    is_list_pair = (
        node.data in [structure.REQUIRED_PAIR, structure.OPTIONAL_PAIR]
        and node.children[1].data == "list"
    )
    if not is_list_pair:
        return False
    list_node = node.children[1]
    return str(list_node.children[0].data) == structure.REFERENCE
